use crate::Mem;
use minifb::{Key, Scale, ScaleMode, Window, WindowOptions};
use std::{
    fs::File,
    io::Read,
    sync::{Arc, Mutex, RwLock},
    thread::sleep,
    time::{Duration, Instant},
};

const FG_COLOR: u32 = 0xFFCC00;
const BG_COLOR: u32 = 0x110500;

pub struct Crtc {
    memory: Arc<Mutex<Mem>>,
    buffer: Vec<u32>,
    window: minifb::Window,
    char_rom: Vec<u8>,
}

pub fn get_char_bin(path: &str) -> Vec<u8> {
    let mut file = File::open(path).unwrap();
    let mut buffer = vec![0; 0xFFFF];
    file.read_exact(&mut buffer).unwrap();
    buffer
}

impl Crtc {
    pub fn new(memory: Arc<Mutex<Mem>>) -> Self {
        let window = Window::new(
            "ʕ·ᴥ·ʔ-☆",
            512,
            380,
            WindowOptions {
                resize: true,
                borderless: true,
                title: true,
                transparency: false,
                scale: Scale::X2,
                scale_mode: ScaleMode::AspectRatioStretch,
                topmost: false,
                none: true,
            },
        )
        .unwrap();
        let char_rom = get_char_bin("./src/cozette.bin");
        Self {
            memory,
            buffer: vec![0; 512 * 380],
            window,
            char_rom,
        }
    }
    fn draw(&mut self) {
        let memory = match self.memory.lock() {
            Ok(memory) => memory,
            Err(_) => {
                println!("Couldn't acquire read on shared memory in video thread");
                return;
            }
        };
        let hw_ctrl = memory.read(0x4000).unwrap();
        match hw_ctrl & 0b0000_1000 == 0b0000_1000 {
            false => {
                // the rest of this function is arcane wizardry based on the specifics of george's weird
                // display and characters... don't fuck around w it
                let mut i = 0;
                for char_row in 0..29 {
                    for char_col in 0..64 {
                        let ascii = memory.read(0x6000 + i).unwrap();
                        i += 1;
                        for row in 0..13 {
                            let byte = self.char_rom[ascii as usize + (row * 0x101)];
                            for i in (0..8).rev() {
                                let buffer_index =
                                    ((char_row) * 13 + (row)) * 512 + (char_col * 8 + i);
                                if (byte << i) & 0x80 == 0x80 {
                                    self.buffer[buffer_index] = FG_COLOR;
                                } else {
                                    self.buffer[buffer_index] = BG_COLOR;
                                }
                            }
                        }
                    }
                }
            }
            true => {
                for addr in 0x6000..0xBF00 {
                    // TODO: eventually this will access memory in the weird interleaved way the hardware is
                    // designed, but this is easiest to implement for now
                    let byte = memory.read(addr).unwrap();
                    for i in 0..8 {
                        match byte & 0x80 >> i == 0 {
                            true => self.buffer[(addr - 0x6000) as usize * 8 + i] = BG_COLOR,
                            false => self.buffer[(addr - 0x6000) as usize * 8 + i] = FG_COLOR,
                        }
                    }
                }
            }
        };

        self.window
            .update_with_buffer(&self.buffer, 512, 380)
            .unwrap();
    }

    pub fn run(&mut self) {
        let frame_duration = Duration::from_millis(16);
        let mut previous_draw = Instant::now();

        loop {
            let mut row0 = 0;
            let mut row1 = 0;
            let mut row2 = 0;
            let mut row3 = 0;
            let mut row4 = 0;
            let mut row5 = 0;

            self.window.get_keys().iter().for_each(|key| match key {
                Key::Escape => row0 ^= 0b1000_0000,
                Key::W => row0 ^= 0b0100_0000,
                Key::E => row0 ^= 0b0010_0000,
                Key::R => row0 ^= 0b0001_0000,
                Key::T => row0 ^= 0b0000_1000,
                Key::U => row0 ^= 0b0000_0100,
                Key::O => row0 ^= 0b0000_0010,
                Key::Backspace => row0 ^= 0b0000_0001,
                Key::Tab => row1 ^= 0b1000_0000,
                Key::Q => row1 ^= 0b0100_0000,
                Key::S => row1 ^= 0b0010_0000,
                Key::G => row1 ^= 0b0001_0000,
                Key::Y => row1 ^= 0b0000_1000,
                Key::I => row1 ^= 0b0000_0100,
                Key::P => row1 ^= 0b0000_0010,
                Key::Enter => row1 ^= 0b0000_0001,
                Key::LeftShift | Key::RightShift => row2 ^= 0b1000_0000,
                Key::D => row2 ^= 0b0100_0000,
                Key::V => row2 ^= 0b0010_0000,
                Key::H => row2 ^= 0b0001_0000,
                Key::K => row2 ^= 0b0000_1000,
                Key::Apostrophe => row2 ^= 0b0000_0100,
                Key::Slash => row2 ^= 0b0000_0010,
                Key::A => row2 ^= 0b0000_0001,
                Key::LeftCtrl | Key::RightCtrl => row3 ^= 0b1000_0000,
                Key::Z => row3 ^= 0b0100_0000,
                Key::F => row3 ^= 0b0010_0000,
                Key::B => row3 ^= 0b0001_0000,
                Key::J => row3 ^= 0b0000_1000,
                Key::L => row3 ^= 0b0000_0100,
                Key::Key2 => row3 ^= 0b0000_0010,
                Key::Key4 => row3 ^= 0b0000_0001,
                Key::LeftAlt | Key::RightAlt => row4 ^= 0b1000_0000,
                Key::X => row4 ^= 0b0100_0000,
                Key::C => row4 ^= 0b0010_0000,
                Key::N => row4 ^= 0b0001_0000,
                Key::M => row4 ^= 0b0000_1000,
                Key::Comma => row4 ^= 0b0000_0100,
                Key::Key1 => row4 ^= 0b0000_0010,
                Key::Key3 => row4 ^= 0b0000_0001,
                Key::LeftSuper => row5 ^= 0b1000_0000,
                Key::Space => row5 ^= 0b0100_0000,
                Key::RightSuper => row5 ^= 0b0010_0000,
                _ => {}
            });

            match &mut self.memory.lock() {
                Ok(memory) => {
                    memory.write(0x4400, row0).unwrap();
                    memory.write(0x4401, row1).unwrap();
                    memory.write(0x4402, row2).unwrap();
                    memory.write(0x4403, row3).unwrap();
                    memory.write(0x4404, row4).unwrap();
                    memory.write(0x4405, row5).unwrap();
                }
                Err(error) => println!("{error}"),
            }
            let now = Instant::now();
            if now - previous_draw > frame_duration {
                self.draw();
                previous_draw = now;
            }
            sleep(Duration::from_millis(1));
        }
    }
}
